!!ARBfp1.0

#Copyright (c) 2004, Stanford University
#All rights reserved.

PARAM light_parms  = program.local[0];
PARAM scales       = program.local[1];
PARAM lightdir     = program.local[2];
PARAM halfway      = program.local[3];

TEMP sample1;
TEMP sample2;
TEMP normal;
TEMP center;
TEMP diffuse;
TEMP specular;
TEMP classification;
ALIAS temp1 = sample1;
ALIAS temp2 = sample2;
ALIAS temp3 = normal;
ALIAS temp4 = center;
ALIAS temp5 = diffuse;
ALIAS temp6 = specular;

# get my density value
TEX center, fragment.texcoord[1], texture[0], 3D;

# Choose only ONE of the following blocks

#######################################################
# central differencing (slower, but more accurate)
TEX temp1, fragment.texcoord[2], texture[0], 3D;
TEX temp2, fragment.texcoord[4], texture[0], 3D;
TEX temp3, fragment.texcoord[6], texture[0], 3D;
TEX temp4, fragment.texcoord[3], texture[0], 3D;
TEX temp5, fragment.texcoord[5], texture[0], 3D;
TEX temp6, fragment.texcoord[7], texture[0], 3D;
MOV sample1.x, temp1.w;
MOV sample1.y, temp2.w;
MOV sample1.z, temp3.w;
MOV sample2.x, temp4.w;
MOV sample2.y, temp5.w;
MOV sample2.z, temp6.w;
SUB normal, sample2, sample1;
#####################################################

#####################################################
# forward differencing 
# (faster, but can cause noise in normals)
#TEX sample1, fragment.texcoord[2], texture[0], 3D;
#MOV normal.x, sample1.w;
#TEX sample1, fragment.texcoord[4], texture[0], 3D;
#MOV normal.y, sample1.w;
#TEX sample1, fragment.texcoord[6], texture[0], 3D;
#MOV normal.z, sample1.w;
#SUB normal, normal, center.w;
####################################################

#gradient mag
DP3 temp1, normal, normal;

#check for gradient overflow (gradient magnitude<=0)
SUB temp1.z, temp1.x, 0.0000000001;
CMP temp1.x, temp1.z, 1.0, temp1.x;

RSQ temp1.x, temp1.x;
RCP temp1.y, temp1.x;

#Base Color
#Build lookup
MUL temp2.x, center.w, scales.x; #map density to 0-1
MUL temp2.y, temp1.y, scales.y;  #map grad_mag to 0-1
TEX classification, temp2, texture[1], 2D;

#Nobj (normalized)
MUL normal, normal, temp1.x;

#N dot L
DP3 diffuse, lightdir, normal;
ABS diffuse, diffuse;

#N dot H
DP3 specular, halfway, normal;
ABS specular, specular; # not needed if shininess is even...

# this forces the shininess param to be 2!!!
# We do this because pow is VERY expensive
MUL specular, specular, specular; 

# add in all the lighting
#temp = classification*light_params.x +
#       classification*light_params.y*diffuse +
#       light_params.z*specular;
MUL temp1, classification, light_parms.y;
MUL temp1, temp1, diffuse;
MAD_SAT temp1, classification, light_parms.x, temp1;
MAD_SAT temp1, light_parms.z, specular, temp1;

#fix alpha
MOV temp1.w, classification.w;

MOV result.color, temp1;

END

# End of program

