/**
 **
 **	This is a simple starter file for a basic parallel volume renderer
 **     using CRUT and Chromium.
 **
 **	The left mouse button allows rotation
 **	The middle mouse button allows scaling
 **
 **	Author: Mike Houston
 **
 **/ 

/* Good old standard includes */  
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
    
#ifdef WIN32
#include <windows.h>
const float M_PI = { 3.14159265f };
#endif 

/* GL includes */ 
#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string>
#include "chromium.h"
#include "cr_error.h"
#include "crut_api.h"
#include "crut_clientapi.h"
#include "globals.h"
#include <GL/glut.h>

/* Volume render includes */ 
#include "clip.h"
#include "subvolume.h"
#include "vec.h"
#include "fshader.h"
#include "vshader.h"

    
/**
 ** constants:
 **
 ** NOTE: There are a bunch of good reasons to use const variables
 ** instead of #define's.  However, Visual C++ does not allow a const
 ** variable to be an array size or the case in a switch() statement.
 ** So in the following, all constants are const variables except
 ** those which need to be array sizes or in switch() statements.
 ** Those are #defines.  (Microsoft, please fix this...)
 **
 **/ 
    
#define FALSE 0
#define TRUE  1

/* title of these windows: */ 
const char *WINDOWTITLE = { "Raptor" };

/* the escape key: */ 
    
#define ESCAPE		0x1b
    
/* initial window size:	*/ 
const int INIT_WINDOW_SIZE = { 500 };

/* multiplication factors for input interaction:
  (these are known from previous experience)    */ 
const float ANGFACT = { 1. };
const float SCLFACT = { 0.005f };
const float TRANSFACT = { 0.005f };


/* able to use the left mouse for either rotation or scaling, */ 
/* in case have only a 2-button mouse:	*/ 
    
#define ROTATE		0
#define TRANSLATE	1
    
/* minimum allowable scale factor:					*/ 
const float MINSCALE = { 0.05f };


/* active mouse buttons (or them together):				*/ 
const int LEFT = { 4 };
const int MIDDLE = { 2 };
const int RIGHT = { 1 };


/* which projection:							*/ 
const int ORTHO = { FALSE };
const int PERSP = { TRUE };


/* which button:							*/ 
    
#define RESET		0
#define QUIT		1
    
/* window background color (rgba):					*/ 
const float BACKCOLOR[] = { 0.0, 0.0, 0.0, 0. };


/* color and line width for the axes:					*/ 
const float AXES_COLOR[] = { 1., .5, 0. };
const float AXES_WIDTH = { 3. };


/* # of milliseconds to wait before using higher detail:		*/ 
const int MSWAIT = { 2500 };


/* handy to have around:						*/ 
    
#define FALSE	0
#define TRUE	1
    
/**
** non-constant global variables:
**/ 
int ActiveButton;              /* current button that is down          */
int AxesList;                  /* list to hold the axes                */
int AxesOn;                    /* ON or OFF                            */
int Debug;                     /* ON means print debug info            */
int LeftButton;                /* either ROTATE or SCALE               */
int MainWindow = 0;            /* window id for main graphics window   */
float Scale, Scale2;           /* scaling factors                      */
float TransXYZ[3];
int WhichProjection;           /* ORTHO or PERSP                       */
int Xmouse, Ymouse;            /* mouse values                         */
float Xrot, Yrot;              /* rotation angles in degrees           */
int window_width = 500, window_height = 500; 

/**
 ** Volume data
 **/
unsigned short *data;
static GLuint vol_tex;
static GLuint transferID;
float SlicePerVox;
int   local_width = 0, local_height = 0, local_depth = 0;
float vox_width = 0, vox_height = 0, vox_depth = 0;
float vox_min = 0;
float gradient_scale = 1.0;
float density_scale = 1.0;
int   byte_swap = FALSE;
EyeSlicedCube * volume;
matrix4_t ObjToTex;

/** 
 ** Volume handler
 **/
SubvolumeManager * subvolumeManager = NULL;
                                                      
/**
 ** Shaders
 **/ 
FSHADER lit_volume;
VSHADER vol_vertex;

/**
 ** function prototypes:
 **/ 
void Animate (void);
void Axes (float length);
void Buttons (int);
void Display_Loop (void);
void DoRasterString (float, float, float, char *);
void DoStrokeString (float, float, float, float, char *);
void HsvRgb (float[3], float[3]);
void InitGraphicsCR (void);
void InitGraphicsGLUT (void);
void InitVolume (char*, int[3], int[3]);
void InitXferFunc (char*);
void InitShaders (char*);
void InitLists (void);
void InitExtensionsCR (void);
void InitExtensionsGLUT (void);
void Keyboard (unsigned char, int, int);
void MouseButton (int, int, int, int);
void MouseMotion (int, int);
void Reset (void);
void Resize (int, int);
void Visibility (int);

/* All the extensions we are going to setup.  
   These are defined in globals.h */
glProgramLocalParameter4fvARB_t   glProgramLocalParameter4fvARB_func   = NULL;
glProgramLocalParameter4dvARB_t   glProgramLocalParameter4dvARB_func   = NULL;
glProgramLocalParameter4dARB_t    glProgramLocalParameter4dARB_func    = NULL;
glGetProgramLocalParameterdvARB_t glGetProgramLocalParameterdvARB_func = NULL;
glProgramEnvParameter4fvARB_t     glProgramEnvParameter4fvARB_func     = NULL;
glProgramEnvParameter4fARB_t      glProgramEnvParameter4fARB_func      = NULL;
glProgramEnvParameter4dvARB_t     glProgramEnvParameter4dvARB_func     = NULL;
glProgramEnvParameter4dARB_t      glProgramEnvParameter4dARB_func      = NULL;
glGetProgramEnvParameterdvARB_t   glGetProgramEnvParameterdvARB_func   = NULL;
glGetProgramEnvParameterfvARB_t   glGetProgramEnvParameterfvARB_func   = NULL;
glGenProgramsARB_t                glGenProgramsARB_func                = NULL;
glProgramStringARB_t              glProgramStringARB_func              = NULL;
glBindProgramARB_t                glBindProgramARB_func                = NULL;
glIsProgramARB_t                  glIsProgramARB_func                  = NULL;
glVertexAttrib3fARB_t             glVertexAttrib3fARB_func             = NULL;
glBlendFuncSeparateEXT_t          glBlendFuncSeparateEXT_func          = NULL;
glActiveTextureARB_t              glActiveTextureARB_func              = NULL;
glTexImage3DEXT_t                 glTexImage3DEXT_func                 = NULL;

/* Chromium specific ones for CRUT mode */
glChromiumParametervCRProc    glChromiumParametervCR_func;
glGetChromiumParametervCRProc glGetChromiumParametervCR_func;
crCreateContextProc crCreateContext_func;
crMakeCurrentProc   crMakeCurrent_func;
crSwapBuffersProc   crSwapBuffers_func;


/* By default, assume we are running in single node mode */
static int USE_CRUT = 0;

#ifdef WIN32
#define GETPROCADDRESS(x) wglGetProcAddress(x)
#else
#define GETPROCADDRESS(x) glXGetProcAddressARB((const GLubyte*) x)
#endif


/**
** main program:
**/ 
int  main (int argc, char *argv[]) 
{
	int start[3];
	int end[3];
	if(argc < 10){
		fprintf(stderr, "Usage: %s <volume file> <shader path> <xfer file> <start_x> <start_y> <start_z> <width> <height> <depth> [CR]\n", argv[0]);
		exit(1);
	}
	
	if(argc == 11 && !strncmp(argv[10], "CR", 4)){
		USE_CRUT = 1;
	}
	
	/*
	 *  setup all the graphics stuff:                                
	 */ 
	if(USE_CRUT)
		InitGraphicsCR();
	else
		InitGraphicsGLUT();
	
	/*
	 * setup extensions:
	 */
	if(USE_CRUT)
                InitExtensionsCR();
	else
                InitExtensionsGLUT();
	
	/*
	 * setup volume 
	 */ 
	start[0] = atoi(argv[4]);
	start[1] = atoi(argv[5]);
	start[2] = atoi(argv[6]);
	end[0]   = start[0]+atoi(argv[7]);
	end[1]   = start[1]+atoi(argv[8]);
	end[2]   = start[2]+atoi(argv[9]);
        
    /* Load up volume */
	InitVolume (argv[1], start, end);
    /* Load up the 2D xfer map */
	InitXferFunc (argv[3]);
    
	/*
	 * load shaders 
	 */ 
	InitShaders (argv[2]);
	
	/*
	 * create the display structures that will not change:          
	 */ 
	InitLists ();
	
	/*
	 * init all the global variables used by Display_Loop():             
	 * this will also post a redisplay                              
	 */ 
	Reset ();
	
	/*
	 * draw the scene once and wait for some interaction:           
	 * (will never return)                                          
	 */ 
	if(USE_CRUT)
		crutMainLoop();
	else
		glutMainLoop();
	
	/*
	 * this is here to make the compiler happy:                     
	 */ 
	return 0;
}


/**
 ** this is where one would put code that is to be called
 ** everytime the glut main loop has nothing to do
 **
 ** this is typically where animation parameters are set
 **
 ** do not call Display_Loop() from here -- let glutMainLoop() do it
 **/ 
void  Animate (void) 
{
        
	/*
      put animation stuff in here -- change some global variables  
      for Display_Loop() to find:                                       
	*/ 
	
	/*
      force a call to Display_Loop() next time it is convenient:        
	*/ 
	if(USE_CRUT)
		crutPostRedisplay ();
	else
		glutPostRedisplay();
} 

/**
 ** glui buttons callback:
 **/ 
void  Buttons (int id) 
{
    switch (id)
    {
    case RESET:
        Reset ();
        break;
    case QUIT:
        /*
         * gracefully close out the graphics:                          
         */ 
        glFinish ();
        /*
         * gracefully exit the program:   
         */ 
        exit (0);
    }
}


/**
 ** draw the complete scene:
 **/ 
void  Display_Loop (void) 
{
    int vx, vy, v;             /* viewport dimensions                  */
    int xl, yb;                /* lower-left corner of viewport        */
    
    /*
     * erase the background:                                        
     */ 
    glDrawBuffer (GL_BACK);	
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    /*
     * set the viewport to a square centered in the window:         
     */ 
    vx = window_width;
    vy = window_height;
    
    /* Test getting window size */
    v = vx > vy ? vx : vy;     /* minimum dimension    */
    xl = (vx - v) / 2;
    yb = (vy - v) / 2;
    glViewport (xl, yb, v, v);
    
    /*
     * set the viewing volume:                                      
     * remember that the eye is at the origin looking in -Z         
     * remember that the Z values are actually                      
     * given as DISTANCES IN FRONT OF THE EYE                       
     */ 
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    if (WhichProjection == ORTHO)
        glOrtho (-0.5, 0.5, -0.5, 0.5, 0.01, 100.);
    
    else
        gluPerspective (60., 1., 0.01, 100.);
    
    /*
     * place the object into the viewing volume:                    
     */ 
    glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();
    gluLookAt (0., 0., 1., 0., 0., 0., 0., 1., 0.);
    
    /*
     * translate the object: 
     */ 
    glTranslatef (TransXYZ[0], TransXYZ[1], TransXYZ[2]);
    
    /*
     * rotate the scene:                                            
     */ 
    glRotatef (Yrot, 0., 1., 0.);
    glRotatef (Xrot, 1., 0., 0.);
    
    glColor3f(1.0, 1.0, 1.0);
    
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_3D_EXT);
    glEnable(GL_BLEND);
    
    glBlendFuncSeparateEXT_func(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, 
                                GL_ONE, GL_ONE);
    
    vol_vertex.Bind ();
    
    //Step X in for gradient
    glProgramLocalParameter4dARB_func(GL_VERTEX_PROGRAM_ARB, 0,
                                      1./((float)SlicePerVox * local_width * vox_width/vox_min), 
                                      0., 0., 1.0); 	
    //Step Y in for gradient
    glProgramLocalParameter4dARB_func(GL_VERTEX_PROGRAM_ARB, 1,
                                      0., 
                                      1./((float)SlicePerVox * local_height * vox_height/vox_min), 
                                      0., 1.0); 
    //Step Z in for gradient
    glProgramLocalParameter4dARB_func(GL_VERTEX_PROGRAM_ARB, 2,
                                      0., 0., 
                                      1./((float)SlicePerVox * local_depth * vox_depth/vox_min), 
                                      1.0); 	
    
    
    lit_volume.Bind();	
    
    //Light parms (ambient, diffuse, specular, shininess)
    glProgramLocalParameter4dARB_func(GL_FRAGMENT_PROGRAM_ARB, 0,
                                      0.2, 0.5, 0.0, 1.0); 
    
    //Scaling for xfer function lookup
    //The first paramater scales the gradient_magnitude
    //The second parameter scales the density
    //The third parameter scales the color value
    //The forth parameter adjusts opacity scaling
    glProgramLocalParameter4dARB_func(GL_FRAGMENT_PROGRAM_ARB, 1,
                                      gradient_scale, density_scale, 
                                      1.0, 1.0 ); 
    
	/* we need to transform the light.  This might be faster to do
	 * in a vertex program, but then we have to do it for every
	 * vertex... */
    float light_world_vec[4]; 
    float light_obj_vec[4];
    float half_world_vec[4];
    float half_obj_vec[4];
    
    //set light as headlight
    light_world_vec[0] = light_world_vec[1] = 0.0f;
    light_world_vec[2] = light_world_vec[3] = 1.0f;
    
    matrix4_t modelview, invmodelview;
    
    glGetFloatv (GL_MODELVIEW_MATRIX, (float *) &modelview);
    matrix_invert_4(&invmodelview, &modelview);
    matrix_multiply_vec_4((float*)&light_obj_vec,
                          &invmodelview,
                          (float*)&light_world_vec);
    vec_normalize_3( (float*)&light_obj_vec , (float*)&light_obj_vec );
    
    //light direction in object space
    glProgramLocalParameter4dARB_func(GL_FRAGMENT_PROGRAM_ARB, 2,
                                      light_obj_vec[0], light_obj_vec[1], 
                                      light_obj_vec[2], 1.0 ); 
    
    half_world_vec[0] = light_world_vec[0] + 0.0f;
    half_world_vec[1] = light_world_vec[1] + 0.0f;
    half_world_vec[2] = light_world_vec[2] + 1.0f;
    vec_normalize_3((float*)&half_world_vec,(float*)&half_world_vec);
    half_world_vec[3] = 1.0f;
    
    matrix_multiply_vec_4((float*)&half_obj_vec,
                          &invmodelview,
                          (float*)&half_world_vec);
    
    //half angle direction in object space
    glProgramLocalParameter4dARB_func(GL_FRAGMENT_PROGRAM_ARB, 3,
                                      half_obj_vec[0], half_obj_vec[1], 
                                      half_obj_vec[2], 1.0 );
    
    //Try to bind textures in
    glActiveTextureARB_func(GL_TEXTURE0_ARB);
    glEnable(GL_TEXTURE_3D_EXT);
    glBindTexture(GL_TEXTURE_3D_EXT, vol_tex);
    
    glActiveTextureARB_func(GL_TEXTURE1_ARB);
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, transferID);
    
    volume->render();
    
    /* be nice and hand off the bounding box to Chromium */
    if(USE_CRUT){
        BBox box = volume->getObjSpaceBBox();
        GLfloat bounds[6] = {box.x1, box.y1, box.z1, box.x2, box.y2, box.z2 };
        glChromiumParametervCR_func(GL_OBJECT_BBOX_CR, GL_FLOAT, 6, bounds);
    }
    
    lit_volume.Release();
    vol_vertex.Release();
    
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_3D_EXT);
    
    /*
     * swap the double-buffered framebuffers:                       
     */ 
    if(USE_CRUT)
        crSwapBuffers_func(MainWindow, 0);
    else
        glutSwapBuffers();
}


/**
 ** initialize volume
 **
 **/ 
void  InitVolume (char* data_file, int start[3], int end[3]) 
{
	/* Open the volume file and figure out how big
       all of the dimensions are */
    FILE* dat_file;
    FILE* volume_data;
    int buffersize = 256;
    char* buffer = new char[buffersize];
    char vol_file[256];	
	int i,j;
	int vol_width, vol_height, vol_depth, data_offset;
    
    dat_file = fopen(data_file, "r");
	if(dat_file == NULL){
        perror("WTF:");
		crWarning("Can't open datafile %s", data_file);
		exit(1);
	}
    
    /* defaults */
    vol_width  = 0;
    vol_height = 0;
    vol_depth  = 0;
    vox_width  = 1.0;
    vox_height = 1.0;
    vox_depth  = 1.0;
    byte_swap  = FALSE;
    data_offset = 0;
    
    /* Parse the file (this is a tad hacky...) */
    while(fgets(buffer, 255, dat_file) != NULL){
        char* token;
        if(buffer[0] == '#')
            continue;
        token = strtok(buffer, "\n :\t");
        if(!strcmp(token, "offset")){
            token = strtok(NULL, "\n :\t");
            data_offset = atoi(token);
        }
        else if(!strcmp(token, "datafile")){
            token = strtok(NULL, "\n \t");
            strcpy(vol_file, token);
        }
        else if(!strcmp(token, "width")){
            token = strtok(NULL, "\n :\t");
            vol_width = atoi(token);
        }
        else if(!strcmp(token, "height")){
            token = strtok(NULL, "\n :\t");
            vol_height = atoi(token);
        }
        else if(!strcmp(token, "depth")){
            token = strtok(NULL, "\n :\t");
            vol_depth = atoi(token);
        }
        else if(!strcmp(token, "vox_width")){
            token = strtok(NULL, "\n :\t");
            vox_width = atof(token);
        }
        else if(!strcmp(token, "vox_height")){
            token = strtok(NULL, "\n :\t");
            vox_height = atof(token);
        }
        else if(!strcmp(token, "vox_depth")){
            token = strtok(NULL, "\n :\t");
            vox_depth = atof(token);
        }
        else if(!strcmp(token, "byte_swap")){
            token = strtok(NULL, "\n :\t");
            if(!strcmp(token, "true")){
                byte_swap = TRUE;
            }
            else{
                byte_swap = FALSE;
            }
        }
        else{
            fprintf(stderr, "unknown token: %s\n", token);
        }
    }
    delete buffer;

    if(vol_width  == 0 ||
       vol_height == 0 ||
       vol_depth  == 0){
        
		crWarning("Volume has a dimension of 0, this seems bogus...");
		exit(1);
    }
    
	volume_data = fopen(vol_file, "rb");
	if(volume_data == NULL){
		crWarning("Can't open volume %s", vol_file);
		exit(1);
	}
	
	local_width  = end[0]-start[0];
	local_height = end[1]-start[1];
	local_depth  = end[2]-start[2];
    
	fprintf(stdout, "\'%s\' is a %dx%dx%d volume\n", 
            vol_file, vol_width, vol_height, vol_depth);	
    fprintf(stdout, "Voxel spacing: %2.2fx%2.2fx%2.2f\n", 
            vox_width, vox_height, vox_depth);
	fprintf(stdout, "Loading %dx%dx%d\n", local_width, 
            local_height, local_depth);
    fprintf(stdout, "Byte swaping?");
    if(byte_swap)
        fprintf(stdout, " yes\n");
    else
        fprintf(stdout, " no\n");
    fprintf(stdout, "Data offset: %d bytes\n", data_offset);
    fflush(stdout);
    
	data = new unsigned short[local_width*local_height*local_depth];
    
	for(i=start[2]; i<end[2]; i++){
		for(j=start[1]; j<end[1]; j++){
            long offset=data_offset+((i*vol_height*vol_width+
                                      j*vol_width+start[0])*
                                     sizeof(unsigned short));
            if(fseek(volume_data, offset, SEEK_SET) != 0){
                fprintf(stderr, "seek error\n");
            }
            if(fread(&(data[(i-start[2])*local_height*local_width+
                            (j-start[1])*local_width]), 
                     sizeof(unsigned short), 
                     local_width, volume_data) 
               != (unsigned int)local_width){
                fprintf(stderr, "can't read file from %d\n", (int)offset);
                fprintf(stderr, "i=%d, j=%d\n", i, j);
                exit(1);
            }
		}
		fprintf(stdout, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\bLoaded... %3.2f", 
                100.f*(i-start[2])/(end[2]-start[2]-1));
	}
	fprintf(stdout, "\n");
    fflush(stdout);
    
	unsigned char *data2 = (unsigned char*)data;
	/* Byte swapping */
    if(byte_swap){
        for(i=0; i<local_width*local_height*local_depth; i++){
            unsigned char temp;
            temp = data2[i*2];
            data2[i*2] = data2[i*2+1];
            data2[i*2+1] = temp;
        }
    }
    
	fclose(volume_data);
	
	glEnable(GL_TEXTURE_3D_EXT);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	
	glGenTextures(1, &vol_tex);
	glBindTexture(GL_TEXTURE_3D_EXT, vol_tex);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_R_EXT, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_S,     GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_WRAP_T,     GL_CLAMP);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    
	glTexImage3DEXT_func(GL_TEXTURE_3D_EXT, 0, GL_ALPHA16, 
                         local_width, local_height, local_depth, 
                         0, GL_ALPHA, GL_UNSIGNED_SHORT, data);
	fprintf(stdout, "Texture Loaded\n");
	delete data;
	
    vox_min = vox_width < vox_height ? vox_width:vox_height;
    vox_min = vox_min < vox_depth ? vox_min:vox_depth;
    
	/* if running in parallel, offset world location to match up
	 * with other nodes */
	if(USE_CRUT){
        subvolumeManager =  new 
            SubvolumeManager( vol_width, vol_height, vol_depth, 
                              vox_width, vox_height, vox_depth, 
                              start[0], start[1], start[2], 
                              local_width, 
                              local_height, 
                              local_depth,
                              1,
                              true);
	}
	/* center the volume chunk the user would like */
	else{ 
        subvolumeManager =  new 
            SubvolumeManager( local_width, local_height, local_depth, 
                              vox_width, vox_height, vox_depth, 
                              0, 0, 0, 
                              local_width, 
                              local_height, 
                              local_depth,
                              0,
                              true);
	}
	
	subvolumeManager->setNumSlicesPerVoxel(1.0);
	ObjToTex = subvolumeManager->getObjToTexTransform();     
	
    // ?? Is it faster to do this or to setup lots of local
	// ?? params?
	glMatrixMode(GL_MATRIX0_ARB);
	glLoadMatrixf((float*)&ObjToTex);
	
	volume = subvolumeManager->getResamplingCube ();
	if(USE_CRUT){
		BBox box = volume->getObjSpaceBBox();
		GLfloat bounds[6] = {box.x1, box.y1, box.z1, 
                             box.x2, box.y2, box.z2 };
		glChromiumParametervCR_func(GL_OBJECT_BBOX_CR, GL_FLOAT, 6, bounds);
	}
    
}

void InitXferFunc (char* xfer_file){
    int width;
	int height;
	unsigned char* transfer;
    int buffersize = 256;
    char* buffer = new char[buffersize];
    char data_file[256];
    FILE* data;
	FILE* dat_file = fopen(xfer_file, "r");
    
	if(dat_file == NULL){
		crWarning("Can't open transfer function description file: %s!", xfer_file);
		exit(1);
	}
    
    /* defaults */
    width  = 0;
    height = 0;
    density_scale  = 1.0;
    gradient_scale = 1.0;
    
    /* Parse the file (this is a tad hacky...) */
    while(fgets(buffer, 255, dat_file)){
        char* token;
        if(buffer[0] == '#')
            continue;
        token = strtok(buffer, "\n :\t");
        if(!strcmp(token, "datafile")){
            token = strtok(NULL, "\n \t");
            strcpy(data_file, token);
        }
        else if(!strcmp(token, "width")){
            token = strtok(NULL, "\n :\t");
            width = atoi(token);
        }
        else if(!strcmp(token, "height")){
            token = strtok(NULL, "\n :\t");
            height = atoi(token);
        }
        else if(!strcmp(token, "density_scale")){
            token = strtok(NULL, "\n :\t");
            density_scale = atof(token);
        }
        else if(!strcmp(token, "gradient_scale")){
            token = strtok(NULL, "\n :\t");
            gradient_scale = atof(token);
        }
        else{
            fprintf(stderr, "unknown token: %s\n", token);
        }
    }
    delete buffer;
	fclose(dat_file);
        
    transfer = new unsigned char[width*height*4];
    data = fopen(data_file, "rb");
    if(data == NULL){
		crWarning("Can't open transfer function file: %s!", data_file);
		exit(1);
    }
    fread(transfer, 4, width*height, data);
    fclose(data);
    
    glEnable(GL_TEXTURE_2D);
    glGenTextures(1, &transferID);
    glActiveTextureARB_func(GL_TEXTURE1_ARB);
	glBindTexture(GL_TEXTURE_2D, transferID);
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, width, height, 
                     0, GL_RGBA, GL_UNSIGNED_BYTE, transfer);        
	delete transfer;
} 

void InitExtensionsCR (void){
    /*
     * Get extension function pointers.
     */
    glProgramLocalParameter4fvARB_func = 
        (glProgramLocalParameter4fvARB_t) crGetProcAddress("glProgramLocalParameter4fvARB");
    assert(glProgramLocalParameter4fvARB_func);
    
    glProgramLocalParameter4dvARB_func = 
        (glProgramLocalParameter4dvARB_t) crGetProcAddress("glProgramLocalParameter4dvARB");
    assert(glProgramLocalParameter4dvARB_func);
    
    glProgramLocalParameter4dARB_func = 
        (glProgramLocalParameter4dARB_t) crGetProcAddress("glProgramLocalParameter4dARB");
    assert(glProgramLocalParameter4dARB_func);
    
    glGetProgramLocalParameterdvARB_func = 
        (glGetProgramLocalParameterdvARB_t) crGetProcAddress("glGetProgramLocalParameterdvARB");
    assert(glGetProgramLocalParameterdvARB_func);
    
    glProgramEnvParameter4fvARB_func = 
        (glProgramEnvParameter4fvARB_t) crGetProcAddress("glProgramEnvParameter4fvARB");
    assert(glProgramEnvParameter4fvARB_func);
    
    glProgramEnvParameter4fARB_func = 
        (glProgramEnvParameter4fARB_t) crGetProcAddress("glProgramEnvParameter4fARB");
    assert(glProgramEnvParameter4fARB_func);
    
    glProgramEnvParameter4dvARB_func = 
        (glProgramEnvParameter4dvARB_t) crGetProcAddress("glProgramEnvParameter4dvARB");
    assert(glProgramEnvParameter4dvARB_func);
    
    glProgramEnvParameter4dARB_func = 
        (glProgramEnvParameter4dARB_t) crGetProcAddress("glProgramEnvParameter4dARB");
    assert(glProgramEnvParameter4dARB_func);
    
    glGetProgramEnvParameterdvARB_func = 
        (glGetProgramEnvParameterdvARB_t) crGetProcAddress("glGetProgramEnvParameterdvARB");
    assert(glGetProgramEnvParameterdvARB_func);
    
    glGetProgramEnvParameterfvARB_func = 
        (glGetProgramEnvParameterfvARB_t) crGetProcAddress("glGetProgramEnvParameterfvARB");
    assert(glGetProgramEnvParameterfvARB_func);
    
    glGenProgramsARB_func = 
        (glGenProgramsARB_t) crGetProcAddress("glGenProgramsARB");
    assert(glGenProgramsARB_func);
    
    glProgramStringARB_func = 
        (glProgramStringARB_t) crGetProcAddress("glProgramStringARB");
    assert(glProgramStringARB_func);
    
    glBindProgramARB_func = 
        (glBindProgramARB_t) crGetProcAddress("glBindProgramARB");
    assert(glBindProgramARB_func);
    
    glIsProgramARB_func = 
        (glIsProgramARB_t) crGetProcAddress("glIsProgramARB");
    assert(glIsProgramARB_func);
    
    glVertexAttrib3fARB_func = 
        (glVertexAttrib3fARB_t) crGetProcAddress("glVertexAttrib3fARB");
    assert(glVertexAttrib3fARB_func);
    
    crSwapBuffers_func = 
        (crSwapBuffersProc) crGetProcAddress("crSwapBuffers");
    assert(crSwapBuffers_func);
    
    glGetChromiumParametervCR_func = 
        (glGetChromiumParametervCRProc) crGetProcAddress("glGetChromiumParametervCR");
    assert(glGetChromiumParametervCR_func);
    
    glBlendFuncSeparateEXT_func = 
        (glBlendFuncSeparateEXT_t) crGetProcAddress("glBlendFuncSeparateEXT");
    assert(glBlendFuncSeparateEXT_func);
    
    glActiveTextureARB_func = 
        (glActiveTextureARB_t) crGetProcAddress("glActiveTextureARB");
    assert(glActiveTextureARB_func);
    
    glTexImage3DEXT_func = 
        (glTexImage3DEXT_t) crGetProcAddress("glTexImage3DEXT");
    assert(glTexImage3DEXT_func);
    
    crCreateContext_func = 
        (crCreateContextProc) crGetProcAddress("crCreateContext");
    assert(crCreateContext_func);
    crMakeCurrent_func = 
        (crMakeCurrentProc) crGetProcAddress("crMakeCurrent");
    assert(crMakeCurrent_func);
    crSwapBuffers_func = 
        (crSwapBuffersProc) crGetProcAddress("crSwapBuffers");
    assert(crSwapBuffers_func);
    glChromiumParametervCR_func = 
        (glChromiumParametervCRProc) crGetProcAddress("glChromiumParametervCR");
    assert(glChromiumParametervCR_func);
    glGetChromiumParametervCR_func = 
        (glGetChromiumParametervCRProc) crGetProcAddress("glGetChromiumParametervCR");
    assert(glGetChromiumParametervCR_func);   
}

void InitExtensionsGLUT (void){
    /*
     * Get extension function pointers.
     */
    glProgramLocalParameter4fvARB_func = (glProgramLocalParameter4fvARB_t) GETPROCADDRESS("glProgramLocalParameter4fvARB");
    assert(glProgramLocalParameter4fvARB_func);
    
    glProgramLocalParameter4dvARB_func = (glProgramLocalParameter4dvARB_t) GETPROCADDRESS("glProgramLocalParameter4dvARB");
    assert(glProgramLocalParameter4dvARB_func);
    
    glProgramLocalParameter4dARB_func = (glProgramLocalParameter4dARB_t) GETPROCADDRESS("glProgramLocalParameter4dARB");
    assert(glProgramLocalParameter4dARB_func);
    
    glGetProgramLocalParameterdvARB_func = (glGetProgramLocalParameterdvARB_t) GETPROCADDRESS("glGetProgramLocalParameterdvARB");
    assert(glGetProgramLocalParameterdvARB_func);
    
    glProgramEnvParameter4fvARB_func = (glProgramEnvParameter4fvARB_t) GETPROCADDRESS("glProgramEnvParameter4fvARB");
    assert(glProgramEnvParameter4fvARB_func);
    
    glProgramEnvParameter4fARB_func = (glProgramEnvParameter4fARB_t) GETPROCADDRESS("glProgramEnvParameter4fARB");
    assert(glProgramEnvParameter4fARB_func);
    
    glProgramEnvParameter4dvARB_func = (glProgramEnvParameter4dvARB_t) GETPROCADDRESS("glProgramEnvParameter4dvARB");
    assert(glProgramEnvParameter4dvARB_func);
    
    glProgramEnvParameter4dARB_func = (glProgramEnvParameter4dARB_t) GETPROCADDRESS("glProgramEnvParameter4dARB");
    assert(glProgramEnvParameter4dARB_func);
    
    glGetProgramEnvParameterdvARB_func = (glGetProgramEnvParameterdvARB_t) GETPROCADDRESS("glGetProgramEnvParameterdvARB");
    assert(glGetProgramEnvParameterdvARB_func);
    
    glGetProgramEnvParameterfvARB_func = (glGetProgramEnvParameterfvARB_t) GETPROCADDRESS("glGetProgramEnvParameterfvARB");
    assert(glGetProgramEnvParameterfvARB_func);
    
    glGenProgramsARB_func = (glGenProgramsARB_t) GETPROCADDRESS("glGenProgramsARB");
    assert(glGenProgramsARB_func);
    
    glProgramStringARB_func = (glProgramStringARB_t) GETPROCADDRESS("glProgramStringARB");
    assert(glProgramStringARB_func);
    
    glBindProgramARB_func = (glBindProgramARB_t) GETPROCADDRESS("glBindProgramARB");
    assert(glBindProgramARB_func);
    
    glIsProgramARB_func = (glIsProgramARB_t) GETPROCADDRESS("glIsProgramARB");
    assert(glIsProgramARB_func);
    
    glVertexAttrib3fARB_func = (glVertexAttrib3fARB_t) GETPROCADDRESS("glVertexAttrib3fARB");
    assert(glVertexAttrib3fARB_func);
    
    glBlendFuncSeparateEXT_func = (glBlendFuncSeparateEXT_t) GETPROCADDRESS("glBlendFuncSeparateEXT");
    assert(glBlendFuncSeparateEXT_func);
    
	glActiveTextureARB_func = 
        (glActiveTextureARB_t) GETPROCADDRESS("glActiveTextureARB");
    assert(glActiveTextureARB_func);
    
	glTexImage3DEXT_func = 
        (glTexImage3DEXT_t) GETPROCADDRESS("glTexImage3DEXT");
    assert(glTexImage3DEXT_func);
    
}


void  InitShaders (char* path) 
{
    std::string prog_path;
    char temp[256];
    sprintf(temp, "%s/%s", path, "vol_vertex.vp");
    prog_path = temp;
    vol_vertex.LoadFromFile (prog_path.c_str ());      /* load vertex program */
    sprintf(temp, "%s/%s", path, "lit_volume.fp");
    prog_path = temp;
    lit_volume.LoadFromFile (prog_path.c_str ());      /* load fragment program */
} 

/**
 ** initialize the crut and OpenGL libraries:
 **	also setup display lists and callback functions
 **/ 
void  InitGraphicsCR (void) 
{
	int ctx;
	int visual = CR_RGB_BIT | CR_ALPHA_BIT | CR_DEPTH_BIT | CR_DOUBLE_BIT;
	
	if (Debug)
        fprintf (stderr, "InitGraphicsCR\n");
    
    /* we need to set these up to use them... */
    crMakeCurrent_func = 
        (crMakeCurrentProc) crGetProcAddress("crMakeCurrent");
    assert(crMakeCurrent_func);
	
	/*
      setup the display mode:  
      ( *must* be done before call to crutMakeCurrent) ) 
      ask for color, alpha, double-buffering, and z-buffering:            
	*/ 
	ctx = crutCreateContext(visual);
    
	/*
      setup the callback routines:  
      
      DisplayFunc -- redraw the window   
      ReshapeFunc -- handle the user resizing the window  
      KeyboardFunc -- handle a keyboard input     
      MouseFunc -- handle the mouse button going down or up 
      MotionFunc -- handle the mouse moving with a button down 
      PassiveMotionFunc -- handle the mouse moving with a button up
      VisibilityFunc -- handle a change in window visibility  
      EntryFunc    -- handle the cursor entering or leaving the window 
      SpecialFunc -- handle special keys on the keyboard  
      SpaceballMotionFunc -- handle spaceball translation 
      SpaceballRotateFunc -- handle spaceball rotation
      SpaceballButtonFunc -- handle spaceball button hits
      ButtonBoxFunc -- handle button box hits  
      DialsFunc -- handle dial rotations      
      TabletMotionFunc -- handle digitizing tablet motion 
      TabletButtonFunc -- handle digitizing tablet button hits
      MenuStateFunc -- declare when a pop-up menu is in use        
      IdleFunc -- what to do when nothing else is going on  
      TimerFunc -- trigger something to happen every so often      
	*/ 
	crutDisplayFunc (Display_Loop);
	crutReshapeFunc (Resize);
	crutKeyboardFunc (Keyboard);
	crutVisibilityFunc (Visibility);
	crutMouseFunc (MouseButton);
	crutMotionFunc (MouseMotion);
	crutInitClient();
	
    crMakeCurrent_func(MainWindow, ctx);
}

/**
 ** initialize the crut and OpenGL libraries:
 **	also setup display lists and callback functions
 **/ 
void  InitGraphicsGLUT (void) 
{
	glutInitDisplayMode (GLUT_RGBA | GLUT_ALPHA | GLUT_DOUBLE | GLUT_DEPTH);
	
	if (Debug)
		fprintf (stderr, "InitGraphicsGLUT\n");
	
	/*
      set the initial window configuration:                        
	*/ 
	glutInitWindowSize (INIT_WINDOW_SIZE, INIT_WINDOW_SIZE);
	
	/*
      open the window and set its title:                           
	*/ 
	MainWindow = glutCreateWindow (WINDOWTITLE);
	glutSetWindowTitle (WINDOWTITLE);
	
	/*
      setup the callback routines:  
      
      DisplayFunc -- redraw the window   
      ReshapeFunc -- handle the user resizing the window  
      KeyboardFunc -- handle a keyboard input     
      MouseFunc -- handle the mouse button going down or up 
      MotionFunc -- handle the mouse moving with a button down 
      PassiveMotionFunc -- handle the mouse moving with a button up
      VisibilityFunc -- handle a change in window visibility  
      EntryFunc    -- handle the cursor entering or leaving the window 
      SpecialFunc -- handle special keys on the keyboard  
      SpaceballMotionFunc -- handle spaceball translation 
      SpaceballRotateFunc -- handle spaceball rotation
      SpaceballButtonFunc -- handle spaceball button hits
      ButtonBoxFunc -- handle button box hits  
      DialsFunc -- handle dial rotations      
      TabletMotionFunc -- handle digitizing tablet motion 
      TabletButtonFunc -- handle digitizing tablet button hits
      MenuStateFunc -- declare when a pop-up menu is in use        
      IdleFunc -- what to do when nothing else is going on  
      TimerFunc -- trigger something to happen every so often      
	*/ 
	glutSetWindow (MainWindow);
	glutDisplayFunc (Display_Loop);
	glutReshapeFunc (Resize);
	glutKeyboardFunc (Keyboard);
	glutMouseFunc (MouseButton);
	glutMotionFunc (MouseMotion);
	glutVisibilityFunc (Visibility);
	
	glutTimerFunc(0, NULL, 0);
}



/**
 ** initialize the display lists that will not change:
 **/ 
void  InitLists (void) 
{
    if (Debug)
        fprintf (stderr, "InitLists\n");
    
    /*
      create the axes:                                             
    */ 
    AxesList = glGenLists (1);
    glNewList (AxesList, GL_COMPILE);
    glColor3fv (AXES_COLOR);
    glLineWidth (AXES_WIDTH);
    Axes (1.5);
    glLineWidth (1.);
    glEndList ();
}


/**
 ** the keyboard callback:
 **/ 
void  Keyboard (unsigned char c, int x, int y) 
{
	if (Debug)
        fprintf (stderr, "DoKeyboard: '%c' (0x%0x)\n", c, c);
	switch (c)
	{
	case 'o':
	case 'O':
        WhichProjection = ORTHO;
        break;
	case 'p':
	case 'P':
        WhichProjection = PERSP;
        break;
	case 'q':
	case 'Q':
	case ESCAPE:
        Buttons (QUIT);        /* will not return here         */
        break;                 /* happy compiler               */
	case ',':
	case '<':
        if(SlicePerVox <= 1.0)
			SlicePerVox /= 2.0;	
        else
			SlicePerVox -= 1.0;
        subvolumeManager->setNumSlicesPerVoxel(SlicePerVox);
        ObjToTex = subvolumeManager->getObjToTexTransform();     
		// ?? Is it faster to do this or to setup lots of local
		// ?? params?
		glMatrixMode(GL_MATRIX0_ARB);
		glLoadMatrixf((float*)&ObjToTex);
        
        delete volume;
        volume = subvolumeManager->getResamplingCube ();
        
		if(USE_CRUT)
		{
			BBox box = volume->getObjSpaceBBox();
			GLfloat bounds[6] = {box.x1, box.y1, box.z1, 
                                 box.x2, box.y2, box.z2 };
			glChromiumParametervCR_func(GL_OBJECT_BBOX_CR, 
                                        GL_FLOAT, 6, bounds);
		}
		fprintf(stdout, "%f slices per voxel\n", SlicePerVox);
        fflush(stdout);
        break;
	case '.':
	case '>':
        if(SlicePerVox <= 1.0)
			SlicePerVox *= 2.0;
        else
			SlicePerVox += 1.0;
        subvolumeManager->setNumSlicesPerVoxel(SlicePerVox);
        ObjToTex = subvolumeManager->getObjToTexTransform();    
		// ?? Is it faster to do this or to setup lots of local
		// ?? params?
		glMatrixMode(GL_MATRIX0_ARB);
		glLoadMatrixf((float*)&ObjToTex);
        
        delete volume;
        volume = subvolumeManager->getResamplingCube ();
		if(USE_CRUT)
		{
			BBox box = volume->getObjSpaceBBox();
			GLfloat bounds[6] = {box.x1, box.y1, box.z1, 
                                 box.x2, box.y2, box.z2 };
			glChromiumParametervCR_func(GL_OBJECT_BBOX_CR, GL_FLOAT, 
                                        6, bounds);
		}
		printf("%f slices per voxel\n", SlicePerVox);
        break;
	case 'R':
	case 'r':
        Reset();
        break;
    case 'T':
    case 't':
        if(LeftButton == ROTATE)
            LeftButton = TRANSLATE;
        else
            LeftButton = ROTATE;
        break;
    case 'x':
		Xrot += 10;
		break;
	case 'X':
		Xrot -= 10;
		break;
	case 'y':
		Yrot += 10;
		break;
	case 'Y':
		Yrot -= 10;
		break;
	case 'z':
		Yrot += 10;
		break;
	case 'Z':
		Xrot -= 10;
		break;
	default:
        fprintf (stderr,
                 "Don't know what to do with keyboard hit:: '%c' (0x%0x)\n",
                 c,
                 c);
	}
	/*
      force a call to Display():                                   
	*/ 
	if(USE_CRUT)
		crutPostRedisplay ();
	else
		glutPostRedisplay();
}


/**
 ** called when the mouse button transitions down or up:
 **/ 
void  MouseButton  ( int button, /* GLUT_*_BUTTON                        */ 
                     int state, /* GLUT_UP or GLUT_DOWN                 */ 
                     int x,
                     int y /* where mouse was when button hit      */  )  
{
	int b;            /* LEFT, MIDDLE, or RIGHT               */
	if (Debug)
        fprintf (stderr, "MouseButton: %d, %d, %d, %d\n", 
                 button, state, x, y);
	
	/*
      get the proper button bit mask:                              
	*/ 
	if(USE_CRUT){
		switch (button)
		{
		case CRUT_LEFT_BUTTON:
			b = LEFT;
			break;
		case CRUT_MIDDLE_BUTTON:
			b = MIDDLE;
			break;
		case CRUT_RIGHT_BUTTON:
			b = RIGHT;
			break;
		default:
			b = 0;
			fprintf (stderr, "Unknown mouse button: %d\n", button);
		}
	}
	else{
		switch (button)
		{
		case GLUT_LEFT_BUTTON:
			b = LEFT;
			break;
		case GLUT_MIDDLE_BUTTON:
			b = MIDDLE;
			break;
		case GLUT_RIGHT_BUTTON:
			b = RIGHT;
			break;
		default:
			b = 0;
			fprintf (stderr, "Unknown mouse button: %d\n", button);
		}
	}
        
	
	/*
      button down sets the bit, up clears the bit:                 
	*/ 
	if (state == CRUT_DOWN || state == GLUT_DOWN)
	{
        Xmouse = x;
        Ymouse = y;
        ActiveButton |= b;     /* set the proper bit   */
	}
	
	else
	{
        ActiveButton &= ~b;    /* clear the proper bit */
	}
}


/**
 ** called when the mouse moves while a button is down:
 **/ 
void 
MouseMotion  ( int x, int y /* mouse coords                         */  )  
{
	int dx, dy;                /* change in mouse coordinates          */
	if (Debug)
        fprintf (stderr, "MouseMotion: %d, %d\n", x, y);
	dx = x - Xmouse;         /* change in mouse coords       */
	dy = y - Ymouse;
	if (ActiveButton & LEFT)
	{
        switch (LeftButton)
        {
        case ROTATE:
			Xrot += (ANGFACT * dy);
			Yrot += (ANGFACT * dx);
			break;
        case TRANSLATE: 
            TransXYZ[2] += (TRANSFACT * dy);
            break;
        }
	}
	else if (ActiveButton & MIDDLE)
	{
        TransXYZ[2] += (TRANSFACT * dy);
	}
	else if (ActiveButton & RIGHT)
	{
        TransXYZ[0] += (TRANSFACT * dx);
        TransXYZ[1] += (TRANSFACT * -dy);
	}
	
	Xmouse = x;               /* new current position         */
	Ymouse = y;
	if(USE_CRUT)
		crutPostRedisplay();
	else
		glutPostRedisplay();
}


/**
 ** reset the transformations and the colors:
 **
 ** this only sets the global variables --
 ** the glut main loop is responsible for redrawing the scene
 **/ 
void  Reset (void) 
{
	ActiveButton = 0;
	AxesOn = TRUE;
	Debug = FALSE;
	LeftButton = ROTATE;
	Scale = 1.0;
	Scale2 = 0.0;              /* because add 1. to it in Display()    */
	WhichProjection = PERSP;
	Xrot = Yrot = 0.;
	SlicePerVox = 1.0;
	TransXYZ[0] = TransXYZ[1] = TransXYZ[2] = 0.;    
	/*
      setup the clear values:                                      
	*/ 
	glClearColor (BACKCOLOR[0], BACKCOLOR[1], BACKCOLOR[2], BACKCOLOR[3]);
    
	if(USE_CRUT)
		crutPostRedisplay ();
	else
		glutPostRedisplay();
} 

/**
 ** called when user resizes the window:
 **/ 
void  Resize (int width, int height) 
{
	int winsize[2];
	winsize[0] = width;
	winsize[1] = height;
	if(Debug)
        fprintf (stderr, "ReSize: %d, %d\n", width, height);
	window_width = width;
	window_height = height;
	if(USE_CRUT)
		glChromiumParametervCR_func(GL_WINDOW_SIZE_CR, GL_INT, 2, winsize);
	if(USE_CRUT)
		crutPostRedisplay ();
	else
		glutPostRedisplay();
}

/**
 ** handle a change to the window's visibility:
 **/ 
void  Visibility  ( int state /* GLUT_VISIBLE or GLUT_NOT_VISIBLE     */  )  {
    if (Debug)
        fprintf (stderr, "Visibility: %d\n", state);
    if (state == CRUT_VISIBLE && USE_CRUT)
    {
        crutPostRedisplay ();
    }
    else if (state == GLUT_VISIBLE && !USE_CRUT)
    {
        glutPostRedisplay ();
    }
    else
    {
        
        /*
          could optimize by keeping track of the fact          
          that the window is not visible and avoid             
          animating or redrawing it ...                        
        */ 
    }
}


/* the stroke characters 'X' 'Y' 'Z' :					*/ 
static float xx[] = { 0., 1., 0., 1. };
static float xy[] = { -.5, .5, .5, -.5 };
static float yx[] = { 0.f, 0.f, -.5f, .5f };
static float yy[] = { 0.f, .6f, 1.f, 1.f };
static float zx[] = { 1.f, 0.f, 1.f, 0.f, .25f, .75f };
static float zy[] = { .5f, .5f, -.5f, -.5f, 0.f, 0.f };
static int xorder[] = { 1, 2, -3, 4 };
static int yorder[] = { 1, 2, 3, -2, 4 };
static int zorder[] = { 1, 2, 3, 4, -5, 6 };


/* fraction of the length to use as height of the characters:		*/ 

#define LENFRAC		0.10

/* fraction of length to use as start location of the characters:	*/ 

#define BASEFRAC	1.10

/**
 **	Draw a set of 3D axes:
 **	(length is the axis length in world coordinates)
 **/ 
void  Axes (float length) 
{
	int i, j;                  /* counters                     */
	float fact;                /* character scale factor       */
	float base;                /* character start location     */
	glBegin (GL_LINE_STRIP);
	glVertex3f (length, 0., 0.);
	glVertex3f (0., 0., 0.);
	glVertex3f (0., length, 0.);
	glEnd ();
	glBegin (GL_LINE_STRIP);
	glVertex3f (0., 0., 0.);
	glVertex3f (0., 0., length);
	glEnd ();
	fact = LENFRAC * length;
	base = BASEFRAC * length;
	glBegin (GL_LINE_STRIP);
	for (i = 0; i < 4; i++)
	{
        j = xorder[i];
        if (j < 0)
        {
			glEnd ();
			glBegin (GL_LINE_STRIP);
			j = -j;
        }
        j--;
        glVertex3f (base + fact * xx[j], fact * xy[j], 0.0);
	}
	glEnd ();
	glBegin (GL_LINE_STRIP);
	for (i = 0; i < 5; i++)
	{
        j = yorder[i];
        if (j < 0)
        {
			glEnd ();
			glBegin (GL_LINE_STRIP);
			j = -j;
        }
        j--;
        glVertex3f (fact * yx[j], base + fact * yy[j], 0.0);
	}
	glEnd ();
	glBegin (GL_LINE_STRIP);
	for (i = 0; i < 6; i++)
	{
        j = zorder[i];
        if (j < 0)
        {
			glEnd ();
			glBegin (GL_LINE_STRIP);
			j = -j;
        }
        j--;
        glVertex3f (0.0, fact * zy[j], base + fact * zx[j]);
	}
	glEnd ();
}


/**
 ** routine to convert HSV to RGB
 **
 ** Reference:  Foley, van Dam, Feiner, Hughes,
 **		"Computer Graphics Principles and Practices,"
 **		Additon-Wesley, 1990, pp592-593.
 **	       
 **/ 
void  HsvRgb (float hsv[3], float rgb[3]) 
{
	float h, s, v;             /* hue, sat, value              */
	float r = 0, g = 0, b = 0;             /* red, green, blue             */
	float i, f, p, q, t;       /* interim values               */
	
	/*
      guarantee valid input:                                       
	*/ 
	h = hsv[0] / 60.;
	while (h >= 6.)
        h -= 6.;
	while (h < 0.)
        h += 6.;
	s = hsv[1];
	if (s < 0.)
        s = 0.;
	if (s > 1.)
        s = 1.;
	v = hsv[2];
	if (v < 0.)
        v = 0.;
	if (v > 1.)
        v = 1.;
	
	/*
      if sat==0, then is a gray:                                   
	*/ 
	if (s == 0.0)
	{
        rgb[0] = rgb[1] = rgb[2] = v;
        return;
	}
	
	/*
      get an rgb from the hue itself:                              
	*/ 
	i = floor (h);
	f = h - i;
	p = v * (1. - s);
	q = v * (1. - s * f);
	t = v * (1. - (s * (1. - f)));
	switch ((int) i)
	{
	case 0:
        r = v;
        g = t;
        b = p;
        break;
	case 1:
        r = q;
        g = v;
        b = p;
        break;
	case 2:
        r = p;
        g = v;
        b = t;
        break;
	case 3:
        r = p;
        g = q;
        b = v;
        break;
	case 4:
        r = t;
        g = p;
        b = v;
        break;
	case 5:
        r = v;
        g = p;
        b = q;
        break;
	}
	rgb[0] = r;
	rgb[1] = g;
	rgb[2] = b;
}
