/* Copyright (c) 2004, Stanford University
   All rights reserved.*/

/*
 *  vec.h
 *
 *  Header file for matrix and vector math routines.
 *
 *  Kekoa Proudfoot
 *  12/20/99
 */

#ifndef VEC_H_INCLUDED
#define VEC_H_INCLUDED

/* Matrix struct */
/* Note: mij is row i, col j.  Matrix is currently column major, like GL. */

typedef struct {
        float m00, m10, m20;
        float m01, m11, m21;
        float m02, m12, m22;
} matrix3_t;

typedef struct {
        float m00, m10, m20, m30;
        float m01, m11, m21, m31;
        float m02, m12, m22, m32;
        float m03, m13, m23, m33;
} matrix4_t;

typedef struct {
        float x, y, z, w;
} quat_t;

extern void vec_normalize_3 (float *p, float *v);

extern void matrix_identity_3 (matrix3_t *p);
extern void matrix_invert_3 (matrix3_t *p, matrix3_t *m);
extern void matrix_multiply_3 (matrix3_t *p, matrix3_t *l, matrix3_t *r);
extern void matrix_multiply_vec_3 (float *p, matrix3_t *l, float *r);
extern void matrix_rotate_3 (matrix3_t *p, float angle, 
                             float x, float y, float z);
extern void matrix_scale_3 (matrix3_t *p, float x, float y, float z);
extern void matrix_transpose_3 (matrix3_t *p, matrix3_t *m);

extern void matrix_affine_4 (matrix3_t *p, matrix4_t *m);
extern void matrix_frustum_4 (matrix4_t *p, float l, float r, 
                              float b, float t, float n, float f);
extern void matrix_identity_4 (matrix4_t *p);
extern void matrix_invert_4 (matrix4_t *p, matrix4_t *m);
extern void matrix_lookat_4 (matrix4_t *p, float ex, float ey, float ez, 
                             float cx, float cy, float cz, 
                             float ux, float uy, float uz);
extern void matrix_multiply_4 (matrix4_t *p, matrix4_t *l, matrix4_t *r);
extern void matrix_multiply_vec_4 (float *p, matrix4_t *l, float *r);
extern void matrix_ortho_4 (matrix4_t *p, float l, float r, float b, 
                            float t, float n, float f);
extern void matrix_rotate_4 (matrix4_t *p, float angle, 
                             float x, float y, float z);
extern void matrix_scale_4 (matrix4_t *p, float x, float y, float z);
extern void matrix_translate_4 (matrix4_t *p, float x, float y, float z);
extern void matrix_transpose_4 (matrix4_t *p, matrix4_t *m);

extern void matrix_perspective_4 (matrix4_t *p, float fov, 
                                  float aspect, float n, float f);

extern void quat_from_angle_axis (quat_t *quat, float angle, float *axis);
extern void quat_to_angle_axis (quat_t *quat, float *angle, float *axis);

#endif /* VEC_H_INCLUDED */
