/* Copyright (c) 2004, Stanford University
   All rights reserved.*/

/************************************************
 * vshader.cpp by mhouston@graphics.stanford.edu
 *
 * handles vertex program stuff
 ************************************************/  
    
#include "vshader.h"
#include "chromium.h"
#include <assert.h>
#include "globals.h"
#include <fstream>
#include <string>
typedef std::string STRING;
typedef std::ifstream IFSTREAM;

extern glProgramLocalParameter4fvARB_t glProgramLocalParameter4fvARB_func;
extern glProgramLocalParameter4dvARB_t glProgramLocalParameter4dvARB_func;
extern glProgramLocalParameter4dARB_t glProgramLocalParameter4dARB_func;
extern glGetProgramLocalParameterdvARB_t glGetProgramLocalParameterdvARB_func;

extern glProgramEnvParameter4fvARB_t glProgramEnvParameter4fvARB_func;
extern glProgramEnvParameter4fARB_t glProgramEnvParameter4fARB_func;
extern glProgramEnvParameter4dvARB_t glProgramEnvParameter4dvARB_func;
extern glProgramEnvParameter4dARB_t glProgramEnvParameter4dARB_func;
extern glGetProgramEnvParameterdvARB_t glGetProgramEnvParameterdvARB_func;
extern glGetProgramEnvParameterfvARB_t glGetProgramEnvParameterfvARB_func;

extern glGenProgramsARB_t glGenProgramsARB_func;
extern glProgramStringARB_t glProgramStringARB_func;
extern glBindProgramARB_t glBindProgramARB_func;
extern glIsProgramARB_t glIsProgramARB_func;

extern glVertexAttrib3fARB_t glVertexAttrib3fARB_func;

VSHADER::~VSHADER () 
{
        Destroy ();
} 

bool VSHADER::LoadFromFile (const char *filename) 
{
        IFSTREAM fin (filename);
        if (!fin)
        {
                fprintf (stderr, "Unable to open %s\n", filename);
                return false;
        }
        STRING shader_str;
        STRING line;
        
        // Probably not efficient, but who cares...
        while (std::getline (fin, line))
        {
                shader_str += line + "\n";
        }
        bool retVal = LoadFromString (shader_str.c_str ());
        if (retVal)
        {
                printf ("Loaded program [%s].\n", filename);
        }
        return retVal;
}

bool VSHADER::LoadFromString (const char *program)
{
	bool bRetValue = false;
	
	/*
          destroy program if it already exists 
	*/ 
	if (m_program_id) {
                Destroy ();
	}
	assert (!m_program_id);
	
	// Generate a valid program ID
	assert (glGenProgramsARB_func);
	glGenProgramsARB_func (1, &m_program_id);
	assert (m_program_id);
	glBindProgramARB_func (GL_VERTEX_PROGRAM_ARB, m_program_id);
	
	// Setup the program string
	GLuint len = strlen (program);
	glProgramStringARB_func (GL_VERTEX_PROGRAM_ARB, 
                                 GL_PROGRAM_FORMAT_ASCII_ARB,
                                 len, (const GLubyte *) program);
	if (glGetError () == GL_NO_ERROR) {
                bRetValue = true;
	}
	
	else {
                GLint ePos;
                glGetIntegerv (GL_PROGRAM_ERROR_POSITION_ARB, &ePos);
                const char *error =
			(const char *) glGetString (GL_PROGRAM_ERROR_STRING_ARB);
                fprintf (stderr, "Shader error: %s\n", error);
                fprintf (stderr,
                         "VSHADER::Load(): Error loading the program at location %d: %.30s\n",
                         ePos, program + ePos);
                
                // this it to match the load_one_program functionality
                exit (1);
	} 
	
	if (glIsProgramARB_func (m_program_id) == 0) {
		bRetValue = false;
	}
	return bRetValue;
}
void VSHADER:: Bind (void) const
{
	if (m_program_id) {               
                // Bind the fragment shader program
                glEnable (GL_VERTEX_PROGRAM_ARB);
                glBindProgramARB_func (GL_VERTEX_PROGRAM_ARB, m_program_id);
	}
}
void VSHADER:: Release (void) const
{
	if (m_program_id) {
                glDisable (GL_VERTEX_PROGRAM_ARB);
	}
}
void VSHADER:: Destroy (void)
{
	/*if (m_program_id) {
                glDeleteProgramsARB_func (1, &m_program_id);
                m_program_id = 0;
                }*/
}
